//ArcBallCamera.cpp
#include <windows.h>
#include <zmouse.h>
#include "ArcBallCamera.h"

ArcBallCamera::ArcBallCamera(void)
: mRadius(100)
, mMinRadius(50)
, mMaxRadius(300)
, mMouseWheelDelta(0)
, mNeedToUpdateFrame(false)
{
	D3DXMatrixIdentity(&mWorldMat);
	D3DXMatrixIdentity(&mViewMat);
}

ArcBallCamera::~ArcBallCamera(void)
{
}

void ArcBallCamera::Reset()
{
	mNeedToUpdateFrame = false;
	D3DXMatrixIdentity(&mWorldMat);
	mViewArcBall.Reset();
}

// Update the scene for every frame
void ArcBallCamera::OnFrameMove()
{
	// No need to handle if no drag since last frame move
	if(!mNeedToUpdateFrame)
		return ;
	mNeedToUpdateFrame = false;

	if(mMouseWheelDelta)
	{
		mRadius -= mMouseWheelDelta * mRadius * 0.1f / 360.0f;

		// Make the radius in range of [mMinRadius, mMaxRadius]
		// This can Prevent the cube became too big or too small
		mRadius = max(mRadius, mMinRadius);
		mRadius = min(mRadius, mMaxRadius);
	}

	// The mouse delta is retrieved IN every WM_MOUSE message and do not accumulate, so clear it after one frame
	mMouseWheelDelta = 0;

	// Get the inverse of the view Arcball's rotation matrix
	D3DXMATRIX rotateMat;
	D3DXMatrixInverse(&rotateMat, NULL, mViewArcBall.GetRotationMatrix());

	// Transform vectors based on camera's rotation matrix
	D3DXVECTOR3 worldUpVector;
	D3DXVECTOR3 loacalUpVector = D3DXVECTOR3(0, 1, 0);
	D3DXVec3TransformCoord(&worldUpVector, &loacalUpVector, &rotateMat);

	D3DXVECTOR3 worldAheadVector;
	D3DXVECTOR3 localAheadVector = D3DXVECTOR3(0, 0, 1);
	D3DXVec3TransformCoord(&worldAheadVector, &localAheadVector, &rotateMat);

	// Update the eye point based on a radius away from the lookAt position
	mEyePos = mLookatPos - worldAheadVector * mRadius;

	// Update the view matrix
	D3DXMatrixLookAtLH(&mViewMat, &mEyePos, &mLookatPos, &worldUpVector);
}

void ArcBallCamera:: ForceRotate()
{
	mNeedToUpdateFrame = true;
	int delta = 1;
	int mousex = mViewArcBall.GetWindowWidth() / 2 - delta;
	int mousey = mViewArcBall.GetWindowHeight() / 2;
	mViewArcBall.OnBegin(mousex, mousey);

	mousex = mViewArcBall.GetWindowWidth() / 2 + delta;
	mViewArcBall.OnMove(mousex, mousey);

	mViewArcBall.OnEnd();
}

// This function is used to handling the mouse message for the view arc ball
// include the right button message, the left button message is handled in the game window message process function
LRESULT ArcBallCamera::HandleMessages(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	// update view arc ball
	if(uMsg == WM_RBUTTONDOWN)
	{
		SetCapture(hWnd);

		mNeedToUpdateFrame = true;
		int mousex = (short)LOWORD(lParam);
		int mousey = (short)HIWORD(lParam);
		mViewArcBall.OnBegin(mousex, mousey);
	}

	// mouse move
	if(uMsg == WM_MOUSEMOVE)
	{
		mNeedToUpdateFrame = true;
		int mousex = (short)LOWORD(lParam);
		int mousey = (short)HIWORD(lParam);
		mViewArcBall.OnMove(mousex, mousey);
	}

	// right button up, terminate view arc ball rotation
	if(uMsg == WM_RBUTTONUP)
	{
		mNeedToUpdateFrame = true;
		mViewArcBall.OnEnd();
		ReleaseCapture();
	}

	// Mouse wheel, zoom in/out
	if(uMsg == WM_MOUSEWHEEL) 
	{
		mNeedToUpdateFrame = true;
		mMouseWheelDelta += (short)HIWORD(wParam);
	}

	return TRUE;
}

void ArcBallCamera::SetViewParams(const D3DXVECTOR3& eypePos, const D3DXVECTOR3& lookatPos, const D3DXVECTOR3& upVector)
{
	mRadius = D3DXVec3Length(&eypePos);
	mEyePos = eypePos;
	mLookatPos = lookatPos;
	mUpVector = upVector;

	D3DXMatrixLookAtLH(&mViewMat, &eypePos, &lookatPos, &upVector);
	mNeedToUpdateFrame = true ;
}

void ArcBallCamera::SetProjParams(float fieldOfView, float aspectRatio, float nearPlane, float farPlane)
{
	D3DXMatrixPerspectiveFovLH(&mProjMat, fieldOfView, aspectRatio, nearPlane, farPlane);
	mNeedToUpdateFrame = true;
}

void ArcBallCamera::SetWindow(int windowWidth, int windowHeight, float arcballRadius)
{
	mViewArcBall.SetWindow(windowWidth, windowHeight, arcballRadius);
}

const D3DXMATRIX ArcBallCamera::GetWorldMatrix() const
{
	return mWorldMat;
}

const D3DXMATRIX ArcBallCamera::GetViewMatrix() const
{
	return mViewMat;
}

const D3DXMATRIX ArcBallCamera::GetProjMatrix() const
{
	return mProjMat;
}

const D3DXVECTOR3 ArcBallCamera::GetEyePoint() const
{
	return mEyePos;
}